@@ -17,7 +17,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: silccipher.c,v 1.24 2002/11/24 16:28:07 priikone Exp $ */
+/* $Id: silccipher.c,v 1.24.2.1 2002/12/05 21:38:17 priikone Exp $ */
 
 #include "silcincludes.h"
 #include "ciphers.h"		/* Includes cipher definitions */
@@ -315,6 +315,11 @@ bool silc_cipher_encrypt(SilcCipher cipher, const unsigned char *src,
 			 unsigned char *dst, SilcUInt32 len, 
 			 unsigned char *iv)
 {
+#ifdef SILC_DEBUG
+  assert((len & (cipher->cipher->block_len - 1)) == 0);
+#endif
+  if (len & (cipher->cipher->block_len - 1))
+    return FALSE;
   return cipher->cipher->encrypt(cipher->context, src, dst, len,
 				 iv ? iv : cipher->iv);
 }
@@ -325,6 +330,13 @@ bool silc_cipher_decrypt(SilcCipher cipher, const unsigned char *src,
 			 unsigned char *dst, SilcUInt32 len, 
 			 unsigned char *iv)
 {
+#if 0 /* Remove to 0.9.11 */
+#ifdef SILC_DEBUG
+  assert((len & (cipher->cipher->block_len - 1)) == 0);
+#endif
+  if (len & (cipher->cipher->block_len - 1))
+    return FALSE;
+#endif
   return cipher->cipher->decrypt(cipher->context, src, dst, len,
 				 iv ? iv : cipher->iv);
 }
